# coding: utf-8

# shells.py - functions for validating user shells
#
# Copyright (C) 2013 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

import ctypes
import ctypes.util
import os
import sys


def list_shells():
    """List the shells from /etc/shells."""
    libc = ctypes.CDLL(ctypes.util.find_library("c"))
    getusershell = libc.getusershell
    getusershell.restype = ctypes.c_char_p
    libc.setusershell()
    while True:
        shell = getusershell()
        if not shell:
            break
        yield shell.decode('utf-8')
    libc.endusershell()


def shellexists(shell):
    """Check if the provided shell exists and is executable."""
    return os.path.isfile(shell) and os.access(shell, os.X_OK)


def check(shell, asroot=False):
    """Check if the specified shell is valid and exit if it isn't."""
    # if the shell is listed in /etc/shells, everything should be OK
    if shell in list_shells():
        return
    # if we are not root, bail out
    if not asroot:
        if not shell:
            # FIXME: print to stderr
            print('%s: empty shell not allowed' % sys.argv[0])
        else:
            # FIXME: print to stderr
            print('%s: %s is an invalid shell' % (sys.argv[0], shell))
        sys.exit(1)
    # warn if something seems wrong
    if not shell:
        # FIXME: print to stderr
        print('%s: Warning: setting empty shell' % sys.argv[0])
    elif not shellexists(shell):
        print('%s: Warning: %s does not exist' % (sys.argv[0], shell))



# coding: utf-8

# cmdline.py - functions for handling command-line options
#
# Copyright (C) 2013-2019 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

import argparse

import constants


version_string = '''
%s
Written by Arthur de Jong.

Copyright (C) 2013-2019 Arthur de Jong
This is free software; see the source for copying conditions.  There is NO
warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
'''.strip() % constants.PACKAGE_STRING


class VersionAction(argparse.Action):

    def __init__(self, option_strings, dest,
                 help='output version information and exit'):
        super(VersionAction, self).__init__(
            option_strings=option_strings,
            dest=argparse.SUPPRESS,
            default=argparse.SUPPRESS,
            nargs=0,
            help=help)

    def __call__(self, parser, namespace, values, option_string=None):
        print(version_string)
        parser.exit()


class ListShellsAction(argparse.Action):

    def __init__(self, option_strings, dest,
                 help='list the shells found in /etc/shells'):
        super(ListShellsAction, self).__init__(
            option_strings=option_strings,
            dest=argparse.SUPPRESS,
            default=argparse.SUPPRESS,
            nargs=0,
            help=help)

    def __call__(self, parser, namespace, values, option_string=None):
        import shells
        for shell in shells.list_shells():
            print(shell)
        parser.exit()



#! /usr/bin/python
# coding: utf-8

# getent.py - program for querying nslcd
#
# Copyright (C) 2013-2019 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

import argparse
import re
import socket
import struct
import sys

import constants
from cmdline import VersionAction
from nslcd import NslcdClient


epilog = '''
supported databases:
  aliases, ethers, group, group.bymember, hosts, hostsv4, hostsv6,
  netgroup, netgroup.norec, networks, networksv4, networksv6, passwd,
  protocols, rpc, services, shadow

Report bugs to <%s>.
'''.strip() % constants.PACKAGE_BUGREPORT

# set up command line parser
parser = argparse.ArgumentParser(
    formatter_class=argparse.RawDescriptionHelpFormatter,
    description='Query information in %s via nslcd.' %
    constants.MODULE_NAME.upper(),
    epilog=epilog)
parser.add_argument('-V', '--version', action=VersionAction)
parser.add_argument('database', metavar='DATABASE',
                    help='any database supported by nslcd')
parser.add_argument('keys', metavar='KEY', nargs='*',
                    help='filter returned database values by key')


def write_aliases(con):
    while con.get_response() == constants.NSLCD_RESULT_BEGIN:
        print('%-16s%s' % (
            con.read_string() + ': ',
            ', '.join(con.read_stringlist())))


def getent_aliases(database, keys=None):
    if not keys:
        write_aliases(NslcdClient(constants.NSLCD_ACTION_ALIAS_ALL))
        return
    for key in keys:
        con = NslcdClient(constants.NSLCD_ACTION_ALIAS_BYNAME)
        con.write_string(key)
        write_aliases(con)


def write_ethers(con):
    while con.get_response() == constants.NSLCD_RESULT_BEGIN:
        name = con.read_string()
        ether = con.read_ether()
        print('%s %s' % (ether, name))


def getent_ethers(database, keys=None):
    if not keys:
        write_ethers(NslcdClient(constants.NSLCD_ACTION_ETHER_ALL))
        return
    for key in keys:
        if re.match('^[0-9a-fA-F]{1,2}(:[0-9a-fA-F]{1,2}){5}$', key):
            con = NslcdClient(constants.NSLCD_ACTION_ETHER_BYETHER)
            con.write_ether(key)
        else:
            con = NslcdClient(constants.NSLCD_ACTION_ETHER_BYNAME)
            con.write_string(key)
        write_ethers(con)


def write_group(con):
    while con.get_response() == constants.NSLCD_RESULT_BEGIN:
        print('%s:%s:%d:%s' % (
            con.read_string(),
            con.read_string(),
            con.read_int32(),
            ','.join(con.read_stringlist())))


def getent_group(database, keys=None):
    if not keys:
        write_group(NslcdClient(constants.NSLCD_ACTION_GROUP_ALL))
        return
    for key in keys:
        if database == 'group.bymember':
            con = NslcdClient(constants.NSLCD_ACTION_GROUP_BYMEMBER)
            con.write_string(key)
        elif re.match(r'^\d+$', key):
            con = NslcdClient(constants.NSLCD_ACTION_GROUP_BYGID)
            con.write_int32(int(key))
        else:
            con = NslcdClient(constants.NSLCD_ACTION_GROUP_BYNAME)
            con.write_string(key)
        write_group(con)


def _get_ipv4(value):
    try:
        return socket.inet_pton(socket.AF_INET, value)
    except socket.error:
        return None


def _get_ipv6(value):
    try:
        return socket.inet_pton(socket.AF_INET6, value)
    except socket.error:
        return None


def _get_af(database):
    if database.endswith('v4'):
        return socket.AF_INET
    elif database.endswith('v6'):
        return socket.AF_INET6
    else:
        return None


def write_hosts(con, db_af):
    while con.get_response() == constants.NSLCD_RESULT_BEGIN:
        names = ' '.join([con.read_string()] + con.read_stringlist())
        for af, address in con.read_addresslist():
            if db_af in (af, None):
                print('%-15s %s' % (address, names))


def getent_hosts(database, keys=None):
    db_af = _get_af(database)
    if not keys:
        write_hosts(NslcdClient(constants.NSLCD_ACTION_HOST_ALL), db_af)
        return
    for key in keys:
        ipv4_addr = _get_ipv4(key)
        ipv6_addr = _get_ipv6(key)
        if ipv4_addr and db_af in (socket.AF_INET, None):
            con = NslcdClient(constants.NSLCD_ACTION_HOST_BYADDR)
            con.write_address(socket.AF_INET, ipv4_addr)
        elif ipv6_addr and db_af in (socket.AF_INET6, None):
            con = NslcdClient(constants.NSLCD_ACTION_HOST_BYADDR)
            con.write_address(socket.AF_INET6, ipv6_addr)
        else:
            con = NslcdClient(constants.NSLCD_ACTION_HOST_BYNAME)
            con.write_string(key)
        write_hosts(con, db_af)


def _read_netgroup(con):
    """Read netgroup name, members and triples from stream."""
    name = con.read_string()
    members = []
    tripples = []
    while True:
        member_type = con.read_int32()
        if member_type == constants.NSLCD_NETGROUP_TYPE_NETGROUP:
            members.append(con.read_string())
        elif member_type == constants.NSLCD_NETGROUP_TYPE_TRIPLE:
            tripples.append((
                con.read_string(), con.read_string(), con.read_string()))
        else:
            break
    return name, members, tripples


def _get_getgroups(con, recurse, netgroups=None):
    if netgroups is None:
        netgroups = {}
    while con.get_response() == constants.NSLCD_RESULT_BEGIN:
        name, members, tripples = _read_netgroup(con)
        if not recurse:
            yield (name, members, tripples)
        else:
            netgroups[name] = None
            for netgroup in members:
                if netgroup not in netgroups:
                    con2 = NslcdClient(constants.NSLCD_ACTION_NETGROUP_BYNAME)
                    con2.write_string(netgroup)
                    all(_get_getgroups(con2, recurse, netgroups))
                if netgroups.get(netgroup, None) is not None:
                    tripples += netgroups[netgroup][1]
            netgroups[name] = (members, tripples)
            yield (name, [], tripples)


def write_netgroup(con, recurse):
    for name, members, tripples in _get_getgroups(con, recurse):
        print('%-15s %s' % (name, ' '.join(
            members +
            ['(%s, %s, %s)' % (host, user, domain)
             for host, user, domain in tripples])))


def getent_netgroup(database, keys=None):
    recurse = database == 'netgroup'
    if not keys:
        write_netgroup(
            NslcdClient(constants.NSLCD_ACTION_NETGROUP_ALL), recurse)
        return
    for key in keys:
        con = NslcdClient(constants.NSLCD_ACTION_NETGROUP_BYNAME)
        con.write_string(key)
        write_netgroup(con, recurse)


def write_networks(con, db_af):
    while con.get_response() == constants.NSLCD_RESULT_BEGIN:
        names = ' '.join([con.read_string()] + con.read_stringlist())
        for af, address in con.read_addresslist():
            if db_af in (af, None):
                print('%-22s %s' % (names, address))


def getent_networks(database, keys=None):
    db_af = _get_af(database)
    if not keys:
        write_networks(NslcdClient(constants.NSLCD_ACTION_NETWORK_ALL), db_af)
        return
    for key in keys:
        ipv4_addr = _get_ipv4(key)
        ipv6_addr = _get_ipv6(key)
        if ipv4_addr and db_af in (socket.AF_INET, None):
            con = NslcdClient(constants.NSLCD_ACTION_NETWORK_BYADDR)
            con.write_address(socket.AF_INET, ipv4_addr)
        elif ipv6_addr and db_af in (socket.AF_INET6, None):
            con = NslcdClient(constants.NSLCD_ACTION_NETWORK_BYADDR)
            con.write_address(socket.AF_INET6, ipv6_addr)
        else:
            con = NslcdClient(constants.NSLCD_ACTION_NETWORK_BYNAME)
            con.write_string(key)
        write_networks(con, db_af)


def write_passwd(con):
    while con.get_response() == constants.NSLCD_RESULT_BEGIN:
        print('%s:%s:%d:%d:%s:%s:%s' % (
            con.read_string(),
            con.read_string(),
            con.read_int32(),
            con.read_int32(),
            con.read_string(),
            con.read_string(),
            con.read_string()))


def getent_passwd(database, keys=None):
    if not keys:
        write_passwd(NslcdClient(constants.NSLCD_ACTION_PASSWD_ALL))
        return
    for key in keys:
        if re.match(r'^\d+$', key):
            con = NslcdClient(constants.NSLCD_ACTION_PASSWD_BYUID)
            con.write_int32(int(key))
        else:
            con = NslcdClient(constants.NSLCD_ACTION_PASSWD_BYNAME)
            con.write_string(key)
        write_passwd(con)


def write_protocols(con):
    while con.get_response() == constants.NSLCD_RESULT_BEGIN:
        name = con.read_string()
        aliases = con.read_stringlist()
        number = con.read_int32()
        print('%-21s %d %s' % (name, number, ' '.join(aliases)))


def getent_protocols(database, keys=None):
    if not keys:
        write_protocols(NslcdClient(constants.NSLCD_ACTION_PROTOCOL_ALL))
        return
    for key in keys:
        if re.match(r'^\d+$', key):
            con = NslcdClient(constants.NSLCD_ACTION_PROTOCOL_BYNUMBER)
            con.write_int32(int(key))
        else:
            con = NslcdClient(constants.NSLCD_ACTION_PROTOCOL_BYNAME)
            con.write_string(key)
        write_protocols(con)


def write_rpc(con):
    while con.get_response() == constants.NSLCD_RESULT_BEGIN:
        name = con.read_string()
        aliases = con.read_stringlist()
        number = con.read_int32()
        print('%-15s %d  %s' % (name, number, ' '.join(aliases)))


def getent_rpc(database, keys=None):
    if not keys:
        write_rpc(NslcdClient(constants.NSLCD_ACTION_RPC_ALL))
        return
    for key in keys:
        if re.match(r'^\d+$', key):
            con = NslcdClient(constants.NSLCD_ACTION_RPC_BYNUMBER)
            con.write_int32(int(key))
        else:
            con = NslcdClient(constants.NSLCD_ACTION_RPC_BYNAME)
            con.write_string(key)
        write_rpc(con)


def write_services(con):
    while con.get_response() == constants.NSLCD_RESULT_BEGIN:
        name = con.read_string()
        aliases = con.read_stringlist()
        number = con.read_int32()
        protocol = con.read_string()
        print('%-21s %d/%s %s' % (name, number, protocol, ' '.join(aliases)))


def getent_services(database, keys=None):
    if not keys:
        write_services(NslcdClient(constants.NSLCD_ACTION_SERVICE_ALL))
        return
    for key in keys:
        value = key
        protocol = ''
        if '/' in value:
            value, protocol = value.split('/', 1)
        if re.match(r'^\d+$', value):
            con = NslcdClient(constants.NSLCD_ACTION_SERVICE_BYNUMBER)
            con.write_int32(int(value))
            con.write_string(protocol)
        else:
            con = NslcdClient(constants.NSLCD_ACTION_SERVICE_BYNAME)
            con.write_string(value)
            con.write_string(protocol)
        write_services(con)


def _shadow_value2str(number):
    return str(number) if number != -1 else ''


def write_shadow(con):
    while con.get_response() == constants.NSLCD_RESULT_BEGIN:
        print('%s:%s:%s:%s:%s:%s:%s:%s:%s' % (
            con.read_string(),
            con.read_string(),
            _shadow_value2str(con.read_int32()),
            _shadow_value2str(con.read_int32()),
            _shadow_value2str(con.read_int32()),
            _shadow_value2str(con.read_int32()),
            _shadow_value2str(con.read_int32()),
            _shadow_value2str(con.read_int32()),
            _shadow_value2str(con.read_int32())))


def getent_shadow(database, keys=None):
    if not keys:
        write_shadow(NslcdClient(constants.NSLCD_ACTION_SHADOW_ALL))
        return
    for key in keys:
        con = NslcdClient(constants.NSLCD_ACTION_SHADOW_BYNAME)
        con.write_string(key)
        write_shadow(con)


def main():  # noqa: C901
    args = parser.parse_args()
    try:
        if args.database == 'aliases':
            getent_aliases(args.database, args.keys)
        elif args.database == 'ethers':
            getent_ethers(args.database, args.keys)
        elif args.database in ('group', 'group.bymember'):
            getent_group(args.database, args.keys)
        elif args.database in ('hosts', 'hostsv4', 'hostsv6'):
            getent_hosts(args.database, args.keys)
        elif args.database in ('netgroup', 'netgroup.norec'):
            getent_netgroup(args.database, args.keys)
        elif args.database in ('networks', 'networksv4', 'networksv6'):
            getent_networks(args.database, args.keys)
        elif args.database == 'passwd':
            getent_passwd(args.database, args.keys)
        elif args.database == 'protocols':
            getent_protocols(args.database, args.keys)
        elif args.database == 'rpc':
            getent_rpc(args.database, args.keys)
        elif args.database == 'services':
            getent_services(args.database, args.keys)
        elif args.database == 'shadow':
            getent_shadow(args.database, args.keys)
        else:
            parser.error('Unknown database: %s' % args.database)
    except struct.error:
        print('Problem communicating with nslcd')
        sys.exit(1)


if __name__ == '__main__':
    main()




# constants.py - configured information and constants, this file is processed
#                and extended by the configure script
#
# Copyright (C) 2012-2015 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA


# Name of package
PACKAGE = '''nss-pam-ldapd'''

# Define to the address where bug reports for this package should be sent.
PACKAGE_BUGREPORT = '''nss-pam-ldapd-users@lists.arthurdejong.org'''

# Define to the full name of this package.
PACKAGE_NAME = '''nss-pam-ldapd'''

# Define to the full name and version of this package.
PACKAGE_STRING = '''nss-pam-ldapd 0.9.13'''

# Define to the home page for this package.
PACKAGE_URL = '''https://arthurdejong.org/nss-pam-ldapd/'''

# Version number of package
VERSION = '''0.9.13'''

# Path to nslcd configuration file.
NSLCD_CONF_PATH = '''/etc/nslcd.conf'''

# The location of the pidfile used for checking availability of the nslcd.
NSLCD_PIDFILE = '''/var/run/nslcd/nslcd.pid'''

# The location of the socket used for communicating.
NSLCD_SOCKET = '''/var/run/nslcd/socket'''

# The SONAME of the NSS library module.
NSS_LDAP_SONAME = '''libnss_ldap.so.2'''

# The name of the NSS and PAM modules.
MODULE_NAME = '''ldap'''

# The following is automatically generated from nslcd.h.
# See that file for details.

NSLCD_VERSION = 0x00000002
NSLCD_ACTION_CONFIG_GET = 0x00010001
NSLCD_CONFIG_PAM_PASSWORD_PROHIBIT_MESSAGE = 1
NSLCD_ACTION_ALIAS_BYNAME = 0x00020001
NSLCD_ACTION_ALIAS_ALL = 0x00020008
NSLCD_ACTION_ETHER_BYNAME = 0x00030001
NSLCD_ACTION_ETHER_BYETHER = 0x00030002
NSLCD_ACTION_ETHER_ALL = 0x00030008
NSLCD_ACTION_GROUP_BYNAME = 0x00040001
NSLCD_ACTION_GROUP_BYGID = 0x00040002
NSLCD_ACTION_GROUP_BYMEMBER = 0x00040006
NSLCD_ACTION_GROUP_ALL = 0x00040008
NSLCD_ACTION_HOST_BYNAME = 0x00050001
NSLCD_ACTION_HOST_BYADDR = 0x00050002
NSLCD_ACTION_HOST_ALL = 0x00050008
NSLCD_ACTION_NETGROUP_BYNAME = 0x00060001
NSLCD_ACTION_NETGROUP_ALL = 0x00060008
NSLCD_NETGROUP_TYPE_NETGROUP = 1
NSLCD_NETGROUP_TYPE_TRIPLE = 2
NSLCD_NETGROUP_TYPE_END = 3
NSLCD_ACTION_NETWORK_BYNAME = 0x00070001
NSLCD_ACTION_NETWORK_BYADDR = 0x00070002
NSLCD_ACTION_NETWORK_ALL = 0x00070008
NSLCD_ACTION_PASSWD_BYNAME = 0x00080001
NSLCD_ACTION_PASSWD_BYUID = 0x00080002
NSLCD_ACTION_PASSWD_ALL = 0x00080008
NSLCD_ACTION_PROTOCOL_BYNAME = 0x00090001
NSLCD_ACTION_PROTOCOL_BYNUMBER = 0x00090002
NSLCD_ACTION_PROTOCOL_ALL = 0x00090008
NSLCD_ACTION_RPC_BYNAME = 0x000a0001
NSLCD_ACTION_RPC_BYNUMBER = 0x000a0002
NSLCD_ACTION_RPC_ALL = 0x000a0008
NSLCD_ACTION_SERVICE_BYNAME = 0x000b0001
NSLCD_ACTION_SERVICE_BYNUMBER = 0x000b0002
NSLCD_ACTION_SERVICE_ALL = 0x000b0008
NSLCD_ACTION_SHADOW_BYNAME = 0x000c0001
NSLCD_ACTION_SHADOW_ALL = 0x000c0008
NSLCD_ACTION_PAM_AUTHC = 0x000d0001
NSLCD_ACTION_PAM_AUTHZ = 0x000d0002
NSLCD_ACTION_PAM_SESS_O = 0x000d0003
NSLCD_ACTION_PAM_SESS_C = 0x000d0004
NSLCD_ACTION_PAM_PWMOD = 0x000d0005
NSLCD_ACTION_USERMOD = 0x000e0001
NSLCD_USERMOD_END = 0
NSLCD_USERMOD_RESULT = 1
NSLCD_USERMOD_FULLNAME = 2
NSLCD_USERMOD_ROOMNUMBER = 3
NSLCD_USERMOD_WORKPHONE = 4
NSLCD_USERMOD_HOMEPHONE = 5
NSLCD_USERMOD_OTHER = 6
NSLCD_USERMOD_HOMEDIR = 7
NSLCD_USERMOD_SHELL = 8
NSLCD_RESULT_BEGIN = 1
NSLCD_RESULT_END = 2
NSLCD_PAM_SUCCESS = 0
NSLCD_PAM_PERM_DENIED = 6
NSLCD_PAM_AUTH_ERR = 7
NSLCD_PAM_CRED_INSUFFICIENT = 8
NSLCD_PAM_AUTHINFO_UNAVAIL = 9
NSLCD_PAM_USER_UNKNOWN = 10
NSLCD_PAM_MAXTRIES = 11
NSLCD_PAM_NEW_AUTHTOK_REQD = 12
NSLCD_PAM_ACCT_EXPIRED = 13
NSLCD_PAM_SESSION_ERR = 14
NSLCD_PAM_AUTHTOK_ERR = 20
NSLCD_PAM_AUTHTOK_DISABLE_AGING = 23
NSLCD_PAM_IGNORE = 25
NSLCD_PAM_ABORT = 26
NSLCD_PAM_AUTHTOK_EXPIRED = 27



# coding: utf-8

# nslcd.py - functions for doing nslcd requests
#
# Copyright (C) 2013-2019 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

import fcntl
import os
import socket
import struct
import sys

import constants


# definition for reading and writing INT32 values
_int32 = struct.Struct('!i')


class NslcdClient(object):

    def __init__(self, action):
        # set up the socket (store in class to avoid closing it)
        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        fcntl.fcntl(self.sock, fcntl.F_SETFD, fcntl.FD_CLOEXEC)
        # connect to nslcd
        self.sock.connect(constants.NSLCD_SOCKET)
        # self.sock.setblocking(1)
        self.fp = os.fdopen(self.sock.fileno(), 'r+b', 0)
        # write a request header with a request code
        self.action = action
        self.write_int32(constants.NSLCD_VERSION)
        self.write_int32(action)

    def write(self, value):
        self.fp.write(value)

    def write_int32(self, value):
        self.write(_int32.pack(value))

    def write_bytes(self, value):
        self.write_int32(len(value))
        if value:
            self.write(value)

    def write_string(self, value):
        if sys.version_info[0] >= 3:
            value = value.encode('utf-8')
        self.write_bytes(value)

    def write_ether(self, value):
        value = struct.pack('BBBBBB', *(int(x, 16) for x in value.split(':')))
        self.write(value)

    def write_address(self, af, value):
        self.write_int32(af)
        self.write_bytes(value)

    def read(self, size):
        value = b''
        while len(value) < size:
            data = self.fp.read(size - len(value))
            if not data:
                raise IOError('NSLCD protocol cut short')
            value += data
        return value

    def read_int32(self):
        return _int32.unpack(self.read(_int32.size))[0]

    def read_bytes(self):
        return self.read(self.read_int32())

    def read_string(self):
        value = self.read_bytes()
        if sys.version_info[0] >= 3:
            value = value.decode('utf-8')
        return value

    def read_stringlist(self):
        num = self.read_int32()
        return [self.read_string() for x in range(num)]

    def read_ether(self):
        value = self.fp.read(6)
        return ':'.join('%x' % x for x in struct.unpack('6B', value))

    def read_address(self):
        af = self.read_int32()
        return af, socket.inet_ntop(af, self.read_bytes())

    def read_addresslist(self):
        num = self.read_int32()
        return [self.read_address() for x in range(num)]

    def get_response(self):
        # complete the request if required and check response header
        if self.action:
            # flush the stream
            self.fp.flush()
            # read and check response version number
            if self.read_int32() != constants.NSLCD_VERSION:
                raise IOError('NSLCD protocol error')
            if self.read_int32() != self.action:
                raise IOError('NSLCD protocol error')
            # reset action to ensure that it is only the first time
            self.action = None
        # get the NSLCD_RESULT_* marker and return it
        return self.read_int32()

    def close(self):
        if hasattr(self, 'fp'):
            try:
                self.fp.close()
            except IOError:
                pass

    def __del__(self):
        self.close()


def usermod(username, asroot=False, password=None, args=None):
    # open a connection to nslcd
    con = NslcdClient(constants.NSLCD_ACTION_USERMOD)
    # write the request information
    con.write_string(username)
    con.write_int32(1 if asroot else 0)
    con.write_string(password)
    for k, v in args.items():
        con.write_int32(k)
        con.write_string(v)
    con.write_int32(constants.NSLCD_USERMOD_END)
    # read the response
    if con.get_response() != constants.NSLCD_RESULT_BEGIN:
        raise IOError('NSLCD protocol error')
    response = {}
    while True:
        key = con.read_int32()
        if key == constants.NSLCD_USERMOD_END:
            break
        response[key] = con.read_string()
    # return the response
    return response



# coding: utf-8

# users.py - functions for validating the user to change information for
#
# Copyright (C) 2013-2019 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

import getpass
import os
import pwd
import sys


class User(object):

    def __init__(self, username):
        self.myuid = os.getuid()
        if username:
            userinfo = pwd.getpwnam(username)
        else:
            self.asroot = False
            userinfo = pwd.getpwuid(self.myuid)
        (self.username, self.password, self.uid, self.gid, self.gecos,
            self.homedir, self.shell) = userinfo
        # if we are trying to modify another user we should be root
        self.asroot = self.myuid != self.uid

    def check(self):
        """Check whether we can modify the user.

        Check if the user is an LDAP user and whether we may modify the user
        information.
        """
        if self.asroot and self.myuid != 0:
            print("%s: you may not modify user '%s'.\n" %
                  (sys.argv[0], self.username))
            sys.exit(1)
        # FIXME: check if the user is an LDAP user

    def get_passwd(self):
        """Ask and return a password that is required to change the user."""
        # FIXME: only ask the password if we require it
        # (e.g. when root and nslcd has userpwmoddn we don't need to)
        return getpass.getpass(
            'LDAP administrator password: '
            if self.asroot else
            'LDAP password for %s: ' % self.username)
        # FIXME: check if the provided password is valid



#! /usr/bin/python
# coding: utf-8

# chsh.py - program for changing the login shell using nslcd
#
# Copyright (C) 2013-2019 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

import argparse

import constants
import nslcd
import shells
import users
from cmdline import ListShellsAction, VersionAction


# set up command line parser
parser = argparse.ArgumentParser(
    description='Change the user login shell in LDAP.',
    epilog='Report bugs to <%s>.' % constants.PACKAGE_BUGREPORT)
parser.add_argument('-V', '--version', action=VersionAction)
parser.add_argument('-s', '--shell', help='login shell for the user account')
parser.add_argument('-l', '--list-shells', action=ListShellsAction)
parser.add_argument('username', metavar='USER', nargs='?',
                    help="the user who's shell to change")


def ask_shell(oldshell):
    """Ask the user to provide a shell."""
    # Provide Python 2 compatibility
    prompt = '  Login Shell [%s]: ' % oldshell
    try:
        shell = raw_input(prompt)
    except NameError:
        shell = input(prompt)
    return shell or oldshell


def main():
    # parse arguments
    args = parser.parse_args()
    # check username part
    user = users.User(args.username)
    user.check()
    # check the command line shell if one was provided (to fail early)
    shell = args.shell
    if shell is not None:
        shells.check(shell, user.asroot)
    # prompt for a password if required
    password = user.get_passwd()
    # prompt for a shell if it was not specified on the command line
    if shell is None:
        print('Enter the new value, or press ENTER for the default')
        shell = ask_shell(user.shell)
        shells.check(shell, user.asroot)
    # perform the modification
    nslcd.usermod(
        user.username, user.asroot, password, {
            constants.NSLCD_USERMOD_SHELL: shell,
        })
    # TODO: print proper response


if __name__ == '__main__':
    main()
